#pragma once




namespace cpphibernate {
namespace mariadb {

    template<typename T_schema>
    driver_t::driver_t(T_schema&& p_schema)
        : _impl         (*this, std::forward<T_schema>(p_schema))
        , _connection   (nullptr)
        { }

    template<typename T_schema>
    driver_t::driver_t(T_schema&& p_schema, ::cppmariadb::connection& p_connection)
        : _impl         (*this, std::forward<T_schema>(p_schema))
        , _connection   (&p_connection)
        { }

    std::ostream& driver_t::print(std::ostream& os) const
        { return _impl.schema->print(os); }

    ::cppmariadb::connection* driver_t::connection() const
        { return _connection; }

    void driver_t::connection(::cppmariadb::connection * p_connection)
        { _connection = p_connection; }

    template<typename... T_args>
    void driver_t::set_filter_inclusive(T_args&&... args)
        { _filter.set_inclusive(_impl.schema, std::forward<T_args>(args)...); }

    template<typename... T_args>
    void driver_t::set_filter_exclusive(T_args&&... args)
        { _filter.set_exclusive(_impl.schema, std::forward<T_args>(args)...); }

    void driver_t::clear_filter()
        { _filter.clear(); }

    auto& driver_t::impl() const
        { return _impl; }

} }
