#pragma once

#include <list>
#include <vector>

#include "container_helper.h"

namespace cpphibernate {
namespace mariadb {

    /* container_helper */

    template<typename T_container>
    struct container_helper<
        T_container,
        mp::enable_if_t<
                mp::is_specialization_of_v<mp::decay_t<T_container>, std::list>
            ||  mp::is_specialization_of_v<mp::decay_t<T_container>, std::vector>
        >>
    {
        using container_type = T_container;
        using value_type     = typename container_type::value_type;

        /**
         * @brief Emplace a new dataset in the container.
         */
        template<typename... X_args>
        static inline value_type& emplace(container_type& container, X_args&&... args)
        {
            container.emplace_back(std::forward<X_args>(args)...);
            return container.back();
        }

        /**
         * @brief Clear the whole container.
         */
        static inline void clear(container_type& container)
        {
            container.clear();
        }
    };

} }
