#pragma once



namespace cpphibernate {
namespace mariadb {

    /**
     * @brief Key type properties for the passed type.
     *
     * @tparam T_key    Type to get key properties for.
     */
    template<typename T_key, typename = void>
    struct key_properties
    {
        using key_type = T_key;

        /**
         * @brief Check if the passed key value is the default value.
         */
        static bool is_default(const key_type& key) = delete;

        /**
         * @brief Returns true if the key type is auto generated, false otherwise.
         */
        static constexpr bool is_auto_generated() = delete;

        /**
         * @brief Returns extra arguments to use for creating the table.
         */
        static constexpr const char * create_table_argument() = delete;

        /**
         * @brief Return the query to create a new key value.
         */
        static constexpr const char * create_key_query() = delete;
    };

} }
