#pragma once

#include <memory>

namespace cpphibernate {
namespace mariadb {

    /**
     * @brief Type properties for the passed type.
     *
     * @tparam T    Type to get properties for.
     */
    template<typename T, typename = void>
    struct type_properties
    {
        /**
         * @brief Get the mariadb type.
         */
        static constexpr const char * type() = delete;

        /**
         * @brief Convert a value from the database to its actual type.
         */
        static T convert_to(const value_t&) = delete;

        /**
         * @brief Convert the actual value to a database value.
         */
        static value_t convert_from(const T&) = delete;

        /**
         * @brief Get the string to start the "convert to" operation.
         */
        static constexpr const char * convert_to_open() = delete;

        /**
         * @brief Get the string to end the "convert to" operation.
         */
        static constexpr const char * convert_to_close() = delete;

        /**
         * @brief Get the string to start the "convert from" operation.
         */
        static constexpr const char * convert_from_open() = delete;

        /**
         * @brief Get the string to end the "convert from" operation.
         */
        static constexpr const char * convert_from_close() = delete;
    };

} }
