#pragma once

#include <cpphibernate/config.h>

namespace cpphibernate
{

    namespace __impl
    {

        template<typename T_lhs, typename T_rhs>
        using equality_compare_equal = mp::bool_t<hana::value(hana::equal(
            hana::type_c<mp::decay_t<T_lhs>>,
            hana::type_c<mp::decay_t<T_rhs>>))>;

        template<typename T_lhs, typename T_rhs>
        using equality_compare_not_equal = mp::bool_t<hana::value(hana::not_equal(
            hana::type_c<mp::decay_t<T_lhs>>,
            hana::type_c<mp::decay_t<T_rhs>>))>;

    }

    /**
     * @brief Tag all objects that are equality comparable to each other.
     *
     * @tparam T_group  Type used for groupig different types.
     */
    template<typename T_group>
    struct tag_equality_comparable
    {
        static constexpr decltype(auto) equality_compare_group = hana::type_c<T_group>;
    };

    /**
     * @brief Equality compare for all objects that has the quality compare tag
     */
    template<typename T_lhs, typename T_rhs>
    constexpr auto operator == (T_lhs&&, T_rhs&&)
        -> mp::enable_if_t<
                mp::is_valid_v<decltype(std::declval<T_lhs>().equality_compare_group)>
            &&  mp::is_valid_v<decltype(std::declval<T_rhs>().equality_compare_group)>
            &&  mp::is_same_v<
                    typename decltype(+std::declval<T_lhs>().equality_compare_group)::type,
                    typename decltype(+std::declval<T_rhs>().equality_compare_group)::type>,
            __impl::equality_compare_equal<T_lhs, T_rhs>>
        { return __impl::equality_compare_equal<T_lhs, T_rhs> { }; }

    /**
     * @brief Equality compare for all objects that has the quality compare tag
     */
    template<typename T_lhs, typename T_rhs>
    constexpr auto operator != (T_lhs&&, T_rhs&&)
        -> mp::enable_if_t<
                mp::is_valid_v<decltype(std::declval<T_lhs>().equality_compare_group)>
            &&  mp::is_valid_v<decltype(std::declval<T_rhs>().equality_compare_group)>
            &&  mp::is_same_v<
                    typename decltype(+std::declval<T_lhs>().equality_compare_group)::type,
                    typename decltype(+std::declval<T_rhs>().equality_compare_group)::type>,
            __impl::equality_compare_not_equal<T_lhs, T_rhs>>
        { return __impl::equality_compare_not_equal<T_lhs, T_rhs> { }; }

}
