#pragma once

#include <cppcore/misc/indent.h>
#include <cppcore/conversion/string.h>

#include "print_container.h"

namespace cpphibernate
{

    namespace __impl
    {

        /* print_container */

        template<typename T_container, typename T_func>
        struct print_container
        {
            using container_type = T_container;
            using func_type      = T_func;

            container_type  container;
            bool            do_indent;
            func_type       func;

            inline void print(std::ostream& os) const
            {
                using namespace ::cppcore;

                auto beg = std::begin(container);
                auto end = std::end  (container);
                if (beg != end)
                {
                    if (do_indent)
                    {
                        os  << indent << "["
                            << incindent;
                    }
                    else
                    {
                        os << "[";
                    }

                    size_t index = 0;
                    for (auto it = beg; it != end; ++it)
                    {
                        if (index++)
                            os << ",";
                        if (!do_indent)
                            os << " ";
                        func(os, *it);
                    }

                    if (do_indent)
                    {
                        os  << decindent
                            << indent << "]";
                    }
                    else
                    {
                        os << " ]";
                    }
                }
                else
                {
                    os << "[ ]";
                }
            }
        };

        /* print_container_builder */

        template<typename X, typename>
        struct print_container_builder
        {
            template<typename... T_args>
            static constexpr decltype(auto) apply(T_args&&...)
                { static_assert(sizeof...(T_args) == -1, "Invalid parameters for make_print_container(...)!"); }
        };

        template<typename T_container, typename T_bool, typename T_func>
        struct print_container_builder<
            mp::list<T_container, T_bool, T_func>,
            mp::enable_if_t<mp::is_same_v<bool, mp::decay_t<T_bool>>>>
        {
            static constexpr decltype(auto) apply(T_container&& container, bool do_indent, T_func&& func)
            {
                using print_container_type = print_container<T_container, T_func>;
                return print_container_type {
                    std::forward<T_container>(container),
                    do_indent,
                    std::forward<T_func>(func),
                };
            }
        };

        template<typename T_container, typename T_bool>
        struct print_container_builder<
            mp::list<T_container, T_bool>,
            mp::enable_if_t<mp::is_same_v<bool, mp::decay_t<T_bool>>>>
        {
            static constexpr decltype(auto) apply(T_container&& container, bool do_indent)
            {
                using namespace ::cppcore;

                return make_print_container(
                    std::forward<T_container>(container),
                    do_indent,
                    [do_indent](auto& os, auto& value)
                    {
                        using namespace ::cppcore;
                        if (do_indent)
                            os << indent;
                        os << to_string(value);
                    });
            }
        };

    }

}

namespace std
{

    template<
        typename T_print_container,
        typename T_char,
        typename T_traits>
    inline auto
        operator << (
            basic_ostream<T_char, T_traits>& os,
            const T_print_container&         ctnr)
        -> ::cppmp::enable_if_t<
            ::cppmp::is_specialization_of_v<
                ::cppmp::decay_t<T_print_container>,
                ::cpphibernate::__impl::print_container>,
            basic_ostream<T_char, T_traits>&>
    {
        ctnr.print(os);
        return os;
    }

}
