#pragma once

#include <cpphibernate/config.h>

namespace cpphibernate
{

    /**
     * @brief Evaluates to the basic dataset type (containers are removed).
     */
    template<typename T>
    struct real_dataset;

    /**
     * @brief Basic dataset type (containers are removed).
     */
    template<typename T>
    using real_dataset_t = typename real_dataset<T>::type;

    /**
     * @brief Extract the type stored in an type definition .
     */
    template<typename T>
    using unwrap_t = typename T::type;

    /**
     * @brief Extract the decayed type in an type definition.
     */
    template<typename T>
    using decay_unwrap_t = typename mp::decay_t<T>::type;

    /**
     * @brief Unwrap the passed hana::type and return an object of the inner type.
     */
    template<typename T>
    constexpr decltype(auto) unwrap(T&& t)
        { return typename decltype(+t)::type { }; }

    /**
     * @brief Get a unique ID for the passed type
     */
    template<typename T_type>
    constexpr decltype(auto) get_type_id(T_type&&);

    /**
     * @brief Evaluates to true_t if the passed type is a container, false_t otherwise.
     */
    template<typename T>
    struct is_container;

    /**
     * @brief Is true if the passed type is a container, false otherwise.
     */
    template<typename T>
    constexpr decltype(auto) is_container_v = is_container<T>::value;

    /**
     * @brief Evaluates to true_t if the passed type is an ordered container type, false_t otherwise.
     */
    template<typename T>
    struct is_ordered;

    /**
     * @brief Is true if the passed type is an ordered container type, false otherwise.
     */
    template<typename T>
    constexpr decltype(auto) is_ordered_v = is_ordered<T>::value;

    /**
     * @brief Evaluates to true_t if the passed type is a nullable type, false_t otherwise.
     */
    template<typename T>
    struct is_pointer;

    /**
     * @brief Is true if the passed type is a pointer type, false otherwise.
     */
    template<typename T>
    constexpr decltype(auto) is_pointer_v = is_pointer<T>::value;

    /**
     * @brief Evaluates to true_t if the passed type is a pointer type, false_t otherwise.
     */
    template<typename T>
    struct is_nullable;

    /**
     * @brief Is true if the passed type is a nullable type, false otherwise.
     */
    template<typename T>
    constexpr decltype(auto) is_nullable_v = is_nullable<T>::value;

}

#include "type_helper.inl"
