#pragma once

#include "type_helper.h"

#include <list>
#include <vector>
#include <memory>

#include <cppcore/misc/type_helper.h>
#include <cppcore/misc/nullable.h>
#include <cpphibernate/config.h>

namespace cpphibernate
{

    namespace __impl
    {

        /* real_dataset_impl */

        template<typename T, typename = void>
        struct real_dataset_impl
            { using type = T; };

        template<typename T>
        struct real_dataset_impl<cppcore::nullable<T>, void>
            { using type = typename real_dataset_impl<T>::type; };

        template<typename T, typename... T_args>
        struct real_dataset_impl<std::unique_ptr<T, T_args...>, void>
            { using type = typename real_dataset_impl<T>::type; };

        template<typename T>
        struct real_dataset_impl<std::shared_ptr<T>, void>
            { using type = typename real_dataset_impl<T>::type; };

        template<typename T, typename... T_args>
        struct real_dataset_impl<std::list<T, T_args...>, void>
            { using type = typename real_dataset_impl<T>::type; };

        template<typename T, typename... T_args>
        struct real_dataset_impl<std::vector<T, T_args...>, void>
            { using type = typename real_dataset_impl<T>::type; };

    }

    /* real_dataset */

    template<typename T>
    struct real_dataset
        : public __impl::real_dataset_impl<T>
        { };

    /* get_dataset_id */

    namespace __impl
    {
        struct counter_type_id
            { };
    }

    template<typename T_type>
    constexpr decltype(auto) get_type_id(T_type&&)
        { return cppcore::get_unique_id<__impl::counter_type_id, mp::decay_t<T_type>>(); }

    /* is_container */

    template<typename T>
    struct is_container
        : public mp::false_t
        { };

    template<typename T, typename... T_args>
    struct is_container<std::list<T, T_args...>>
        : public mp::true_t
        { };

    template<typename T, typename... T_args>
    struct is_container<std::vector<T, T_args...>>
        : public mp::true_t
        { };

    /* is_ordered */

    template<typename T>
    struct is_ordered
        : public mp::false_t
        { };

    template<typename T, typename... T_args>
    struct is_ordered<std::list<T, T_args...>>
        : public mp::true_t
        { };

    template<typename T, typename... T_args>
    struct is_ordered<std::vector<T, T_args...>>
        : public mp::true_t
        { };

    /* is_pointer */

    template<typename T>
    struct is_pointer
        : public mp::false_t
        { };

    template<typename T, typename... T_args>
    struct is_pointer<std::unique_ptr<T, T_args...>>
        : public mp::true_t
        { };

    template<typename T>
    struct is_pointer<std::shared_ptr<T>>
        : public mp::true_t
        { };

    /* is_nullable */

    template<typename T>
    struct is_nullable
        : public is_pointer<T>
        { };

    template<typename T>
    struct is_nullable<cppcore::nullable<T>>
        : public mp::true_t
        { };

}
