#pragma once

#include "limit.h"

namespace cpphibernate
{

    namespace __impl
    {

        /* limit_t */

        template<typename T_value>
        struct limit_t
            : public mp::size_t<T_value::value>
            , public tag_modifier
            , public tag_equality_comparable<tag_modifier>
            { };

        /* limit_builder */

        template<typename X, typename>
        struct limit_builder
        {
            template<typename... T_args>
            static constexpr decltype(auto) apply(T_args&&...)
                { static_assert(sizeof...(T_args) == -1, "Invalid parameters for limit(...)!"); }
        };

        template<typename T>
        struct limit_builder<
            mp::list<T>,
            mp::enable_if_t<
                    mp::is_valid_v<decltype(T::value)>
                &&  mp::is_integral_v<mp::decay_t<decltype(T::value)>>>>
        {
            static constexpr decltype(auto) apply(T&&)
                { return limit_t<T> { }; }
        };

    }

    /* is_limit_modifier */

    template<typename T>
    struct is_limit_modifier
        : public mp::is_specialization_of<T, __impl::limit_t>
        { };

}
