#pragma once

#include "where.h"
#include "modifier.h"
#include "where/where_clause.h"

namespace cpphibernate
{

    namespace __impl
    {

        /* where_t */

        template<typename T_where_clause>
        struct where_t
            : public tag_modifier
            , public tag_equality_comparable<tag_modifier>
        {
            using clause_type = T_where_clause;

            clause_type clause;

            constexpr where_t(clause_type&& p_clause)
                : clause(p_clause)
                { }
        };

        /* where_builder */

        template<typename X, typename>
        struct where_builder
        {
            template<typename... T_args>
            static constexpr decltype(auto) apply(T_args&&...)
                { static_assert(sizeof...(T_args) == -1, "Invalid parameters for where(...)!"); }
        };

        template<typename T_where_clause>
        struct where_builder<
            mp::list<T_where_clause>,
            mp::enable_if_t<mp::is_true_v<is_where_clause_v<mp::decay_t<T_where_clause>>>>>
        {
            static constexpr decltype(auto) apply(T_where_clause&& clause)
            {
                using clause_type = T_where_clause;
                using where_type  = where_t<clause_type>;
                return where_type(std::forward<clause_type>(clause));
            }
        };

    }

    /* is_where_modifier */

    template<typename T>
    struct is_where_modifier
        : public mp::is_specialization_of<T, __impl::where_t>
        { };

}
