#pragma once

#include <cpphibernate/config.h>

namespace cpphibernate
{

    namespace __impl
    {

        /**
         * @brief Helper class to build conjunction objects.
         */
        template<typename X, typename = void>
        struct conjunction_builder;

    }



    /**
     * @brief Evaluates to true_t if the passed type is a conjunction clause, false_t otherwise.
     */
    template<typename T>
    struct is_conjunction;

    /**
     * @brief Is true if the passed type is a conjunction clause, false otherwise.
     */
    template<typename T>
    constexpr decltype(auto) is_conjunction_v = is_conjunction<T>::value;

    /**
     * @brief Predicate to create new conjunction clause objects.
     */
    constexpr decltype(auto) conjunction = mp::generic_predicate<__impl::conjunction_builder> { };



    /**
     * @brief Evaluates to true_t if the passed type is a conjunction clause, false_t otherwise.
     */
    template<typename T>
    struct is_and;

    /**
     * @brief Is true if the passed type is a conjunction clause, false otherwise.
     */
    template<typename T>
    constexpr decltype(auto) is_and_v = is_and<T>::value;

    /**
     * @brief Predicate to create new conjunction clause objects.
     */
    constexpr decltype(auto) and_ = mp::generic_predicate<__impl::conjunction_builder> { };

}

#include "conjunction.inl"
