#pragma once

#include <cpphibernate/schema/field.h>

#include "conjunction.h"

namespace cpphibernate
{

    namespace __impl
    {

        /* where_clause_conjunction_t */

        template<typename... T_clauses>
        struct where_clause_conjunction_t
            : public tag_where_clause
        {
            using clauses_type = hana::basic_tuple<T_clauses...>;

            clauses_type clauses;

            constexpr where_clause_conjunction_t(
                T_clauses&&... p_clauses)
                : clauses(std::forward<T_clauses>(p_clauses)...)
                { }
        };

        /* conjunction_builder */

        template<typename X, typename>
        struct conjunction_builder
        {
            template<typename... T_args>
            static constexpr decltype(auto) apply(T_args&&... args)
                { static_assert(sizeof...(args) == -1, "Invalid parameters for hibernate::conjunction(...)!"); }
        };

        template<typename... T_clauses>
        struct conjunction_builder<
            mp::list<T_clauses...>,
            mp::enable_if_t<
                    mp::is_true_v<is_where_clause_v<mp::decay_t<T_clauses>>...>
                &&  sizeof...(T_clauses) >= 2>>
        {
            static constexpr decltype(auto) apply(T_clauses&&... clauses)
            {
                using clause_type = where_clause_conjunction_t<T_clauses...>;
                return clause_type(std::forward<T_clauses>(clauses)...);
            }
        };

    }

    /* is_conjunction */

    template<typename T>
    struct is_conjunction
        : public mp::is_specialization_of<T, __impl::where_clause_conjunction_t>
        { };

    /* is_and */

    template<typename T>
    struct is_and
        : public mp::is_specialization_of<T, __impl::where_clause_conjunction_t>
        { };

}
