#pragma once

#include <cpphibernate/config.h>

namespace cpphibernate
{

    namespace __impl
    {

        /**
         * @brief Helper class to build disjunction objects.
         */
        template<typename X, typename = void>
        struct disjunction_builder;

    }



    /**
     * @brief Evaluates to true_t if the passed type is a disjunction clause, false_t otherwise.
     */
    template<typename T>
    struct is_disjunction;

    /**
     * @brief Is true if the passed type is a disjunction clause, false otherwise.
     */
    template<typename T>
    constexpr decltype(auto) is_disjunction_v = is_disjunction<T>::value;

    /**
     * @brief Predicate to create new disjunction clause objects.
     */
    constexpr decltype(auto) disjunction = mp::generic_predicate<__impl::disjunction_builder> { };



    /**
     * @brief Evaluates to true_t if the passed type is a disjunction clause, false_t otherwise.
     */
    template<typename T>
    struct is_or;

    /**
     * @brief Is true if the passed type is a disjunction clause, false otherwise.
     */
    template<typename T>
    constexpr decltype(auto) is_or_v = is_or<T>::value;

    /**
     * @brief Predicate to create new disjunction clause objects.
     */
    constexpr decltype(auto) or_ = mp::generic_predicate<__impl::disjunction_builder> { };

}

#include "disjunction.inl"
