#pragma once

#include <cpphibernate/schema/field.h>

#include "equal.h"

namespace cpphibernate
{

    namespace __impl
    {

        /* where_clause_equal_t */

        template<typename T_field, typename T_value>
        struct where_clause_equal_t
            : public tag_where_clause
        {
            using field_type = T_field;
            using value_type = T_value;

            field_type field;
            value_type value;

            constexpr where_clause_equal_t(
                field_type&& p_field,
                value_type&& p_value)
                : field(std::forward<field_type>(p_field))
                , value(std::forward<value_type>(p_value))
                { }
        };

        /* equal_builder */

        template<typename X, typename>
        struct equal_builder
        {
            template<typename... T_args>
            static constexpr decltype(auto) apply(T_args&&... args)
                { static_assert(sizeof...(args) == -1, "Invalid parameters for hibernate::equal(...)!"); }
        };

        template<typename T_field, typename T_value>
        struct equal_builder<
            mp::list<T_field, T_value>,
            mp::enable_if_t<schema::is_field_v<mp::decay_t<T_field>>>>
        {
            static constexpr decltype(auto) apply(T_field&& field, T_value&& value)
            {
                using clause_type = where_clause_equal_t<T_field, T_value>;
                return clause_type(std::forward<T_field>(field), std::forward<T_value>(value));
            }
        };

    }

    /* is_equal */

    template<typename T>
    struct is_equal
        : public mp::is_specialization_of<T, __impl::where_clause_equal_t>
        { };

}
