#pragma once

#include <cpphibernate/config.h>

namespace cpphibernate
{

    namespace __impl
    {

        /**
         * @brief Helper class to build negation objects.
         */
        template<typename X, typename = void>
        struct negation_builder;

    }



    /**
     * @brief Evaluates to true_t if the passed type is a negation clause, false_t otherwise.
     */
    template<typename T>
    struct is_negation;

    /**
     * @brief Is true if the passed type is a negation clause, false otherwise.
     */
    template<typename T>
    constexpr decltype(auto) is_negation_v = is_negation<T>::value;

    /**
     * @brief Predicate to create new negation clause objects.
     */
    constexpr decltype(auto) negation = mp::generic_predicate<__impl::negation_builder> { };



    /**
     * @brief Evaluates to true_t if the passed type is a negation clause, false_t otherwise.
     */
    template<typename T>
    struct is_not;

    /**
     * @brief Is true if the passed type is a negation clause, false otherwise.
     */
    template<typename T>
    constexpr decltype(auto) is_not_v = is_not<T>::value;

    /**
     * @brief Predicate to create new negation clause objects.
     */
    constexpr decltype(auto) not_ = mp::generic_predicate<__impl::negation_builder> { };

}

#include "negation.inl"
