#pragma once

#include <cpphibernate/schema/field.h>

#include "negation.h"

namespace cpphibernate
{

    namespace __impl
    {

        /* where_clause_negation_t */

        template<typename T_clause>
        struct where_clause_negation_t
            : public tag_where_clause
        {
            using clause_type = T_clause;

            clause_type clause;

            constexpr where_clause_negation_t(
                clause_type&& p_clause)
                : clause(std::forward<clause_type>(p_clause))
                { }
        };

        /* negation_builder */

        template<typename X, typename>
        struct negation_builder
        {
            template<typename... T_args>
            static constexpr decltype(auto) apply(T_args&&... args)
                { static_assert(sizeof...(args) == -1, "Invalid parameters for hibernate::negation(...)!"); }
        };

        template<typename T_clause>
        struct negation_builder<
            mp::list<T_clause>,
            mp::enable_if_t<is_where_clause_v<mp::decay_t<T_clause>>>>
        {
            static constexpr decltype(auto) apply(T_clause&& clause)
            {
                using clause_type = where_clause_negation_t<T_clause>;
                return clause_type(std::forward<T_clause>(clause));
            }
        };

    }

    /* is_negation */

    template<typename T>
    struct is_negation
        : public mp::is_specialization_of<T, __impl::where_clause_negation_t>
        { };

    /* is_not */

    template<typename T>
    struct is_not
        : public mp::is_specialization_of<T, __impl::where_clause_negation_t>
        { };

}
