#pragma once

#include <cpphibernate/config.h>
#include <cpphibernate/misc/equality_compare.h>

namespace cpphibernate {
namespace schema {

    namespace __impl
    {

        /**
         * @brief Tag class to mark attribute types.
         */
        struct tag_attribute
            { };

        /**
         * @brief Attribute implementation.
         */
        template<typename T_inner>
        struct attribute_t
            : public T_inner
            , public tag_attribute
            , public tag_equality_comparable<tag_attribute>
        {
            using inner_type = T_inner;
            using this_type  = attribute_t<inner_type>;

/*
            template<typename T_other>
            constexpr decltype(auto) operator==(T_other&&) const
            {
                return ::boost::hana::type<mp::decay_t<decltype(*this)>> { } ==
                    ::boost::hana::type<mp::decay_t<T_other>> { };
            }

            template<typename T_other>
            constexpr decltype(auto) operator!=(T_other&&) const
            {
                return ::boost::hana::type<mp::decay_t<decltype(*this)>> { } !=
                    ::boost::hana::type<mp::decay_t<T_other>> { };
            }
*/
        };

        struct hex_t         { static constexpr decltype(auto) name = "hex"; };
        struct compress_t    { static constexpr decltype(auto) name = "compress"; };
        struct primary_key_t { static constexpr decltype(auto) name = "primary_key"; };

    }

    namespace attribute
    {

        using hex_type          = __impl::attribute_t<__impl::hex_t>;
        using compress_type     = __impl::attribute_t<__impl::compress_t>;
        using primary_key_type  = __impl::attribute_t<__impl::primary_key_t>;

        /**
         * @brief Attribute to indicate that the value is stored as hexadecimal string.
         */
        constexpr hex_type hex { };

        /**
         * @brief Attribute to indicate that the value is stored compressed.
         */
        constexpr compress_type compress { };

        /**
         * @brief Attribute to indicate that the value is the primary key field.
         */
        constexpr primary_key_type primary_key { };

    }

    /**
     * @brief Evaluates to true_t if the passed type is an attribute type.
     */
    template<typename T>
    struct is_attribute;

    /**
     * @brief Is true if the passed type is an attribute, false otherwise.
     */
    template<typename T>
    constexpr bool is_attribute_v = is_attribute<T>::value;

} }

#include "attribute.inl"
