#pragma once

#include "attributes.h"

namespace cpphibernate {
namespace schema {

    namespace __impl
    {

        /* attributes_t */

        template<typename... T_attributes>
        using attributes_t = hana::tuple<T_attributes...>;

        /* attributes_builder */

        template<typename T, typename>
        struct attributes_builder
        {
            template<typename... T_args>
            static constexpr decltype(auto) apply(T_args&&...)
                { static_assert(sizeof...(T_args) == -1, "Invalid parameters for make_attributes(...)!"); }
        };

        template<typename... T>
        struct attributes_builder<
            mp::list<T...>,
            mp::enable_if_t<
                mp::is_true_v<is_attribute_v<mp::decay_t<T>>...>>>
        {
            static constexpr decltype(auto) apply(T&&...)
            {
                using attributes_type = attributes_t<mp::decay_t<T>...>;
                return attributes_type { };
            }
        };

        /* is_attributes_impl */

        template<typename T, typename = void>
        struct is_attributes_impl
            : mp::false_t
            { };

        template<typename... T>
        struct is_attributes_impl<
            attributes_t<T...>,
            mp::enable_if_t<mp::is_true_v<is_attribute_v<mp::decay_t<T>>...>>>
            : mp::true_t
            { };

    }

    /* is_attributes */

    template<typename T>
    struct is_attributes
        : public __impl::is_attributes_impl<T>
        { };

} }
