#pragma once

#include <cppcore/misc/indent.h>
#include <cppcore/misc/type_helper.h>

#include "field.h"
#include "attributes.h"

namespace cpphibernate {
namespace schema {

    namespace __impl
    {

        /* field_t */

        template<typename T_name, typename T_getter, typename T_setter, typename T_attributes>
        constexpr field_t<T_name, T_getter, T_setter, T_attributes>
            ::field_t(
                T_name&&        p_name,
                T_getter&&      p_getter,
                T_setter&&      p_setter,
                T_attributes&&  p_attributes)
                : name          (std::forward<T_name>      (p_name))
                , getter        (std::forward<T_getter>    (p_getter))
                , setter        (std::forward<T_setter>    (p_setter))
                , attributes    (std::forward<T_attributes>(p_attributes))
                { }

        template<typename T_name, typename T_getter, typename T_setter, typename T_attributes>
        void field_t<T_name, T_getter, T_setter, T_attributes>
            ::print(std::ostream& os) const
        {
            using namespace ::cppcore;

            size_t index = 0;
            os  << indent << '{'
                << incindent
                    << indent << "\"name\": \"" << name << "\","
                    << indent << "\"value_type\": \"" << type_helper<value_type>::name() << "\""
                    << indent << "\"dataset_type\": \"" << type_helper<dataset_type>::name() << "\""
                    << indent << "\"attributes\": "
                    << indent << '['
                    << incindent;
                    hana::for_each(attributes, [&](auto& attrib){
                        if (index++ > 0) os << ",";
                        os << indent << '"' << attrib.name << '"';
                    });
            os      << decindent
                    << indent << ']'
                << decindent
                << indent << '}';
        }

        /* field_builder */

        template<typename X, typename>
        struct field_builder
        {
            template<typename... T_args>
            static constexpr decltype(auto) apply(T_args&&...)
                { static_assert(sizeof...(T_args) == -1, "Invalid parameters for make_field(...)!"); }
        };

        template<typename T_name, typename T_getter, typename T_setter, typename T_attributes>
        struct field_builder<
            mp::list<T_name, T_getter, T_setter, T_attributes>,
            mp::enable_if_t<
                    true
                &&  mp::is_getter_v<mp::decay_t<T_getter>>
                &&  mp::is_setter_v<mp::decay_t<T_setter>>
                &&  is_attributes_v<mp::decay_t<T_attributes>>>>
        {
            using name_type         = T_name;
            using getter_type       = T_getter;
            using setter_type       = T_setter;
            using attributes_type   = T_attributes;
            using field_type        = field_t<name_type, getter_type, setter_type, attributes_type>;

            static constexpr decltype(auto) apply(
                name_type&&         name,
                getter_type&&       getter,
                setter_type&&       setter,
                attributes_type&&   attributes)
            {
                return field_type(
                    std::forward<name_type>         (name),
                    std::forward<getter_type>       (getter),
                    std::forward<setter_type>       (setter),
                    std::forward<attributes_type>   (attributes));
            }
        };

    }

    /* is_field */

    template<typename T>
    struct is_field
        : public mp::is_base_of<__impl::tag_field, T>
        { };

} }
