#pragma once

#include "fields.h"

namespace cpphibernate {
namespace schema {

    namespace __impl
    {

        /* fields_t */

        template<typename... T_fields>
        using fields_t = hana::tuple<T_fields...>;

        /* fields_builder */

        template<typename T, typename>
        struct fields_builder
        {
            template<typename... T_args>
            static constexpr decltype(auto) apply(T_args&&...)
                { static_assert(sizeof...(T_args) == -1, "Invalid parameters for make_fields(...)!"); }
        };

        template<typename... T>
        struct fields_builder<
            mp::list<T...>,
            mp::enable_if_t<mp::is_true_v<
                is_field_v<mp::decay_t<T>>...>>>
        {
            template<typename... T_args>
            static constexpr decltype(auto) apply(T_args&&... args)
            {
                using fields_type = fields_t<T_args...>;
                return fields_type(std::forward<T_args>(args)...);
            }
        };

        /* is_fields_impl */

        template<typename T, typename = void>
        struct is_fields_impl
            : mp::false_t
            { };

        template<typename... T>
        struct is_fields_impl<
            fields_t<T...>,
            mp::enable_if_t<mp::is_true_v<is_field_v<mp::decay_t<T>>...>>>
            : mp::true_t
            { };

    }

    /* is_fields */

    template<typename T>
    struct is_fields
        : public __impl::is_fields_impl<T>
        { };

} }
